
import re
import sys
import copy
from matplotlib import pyplot as plt
import numpy as np
import networkx as nx
import pyexcel
import gurobipy as gp
from gurobipy import GRB

G = nx.DiGraph()


iterate = np.array([0, 3.4, 2.6, 2.6, 2.6, 2.6, 2.6, 2.6, 2.6, 2.6, 2.6, 2.6, 2.6, 2.6, 2.6, 2.6, 3.8])
fasit = np.add.accumulate(iterate)
fasit = fasit[:-1]




distanceMatrSimple = np.zeros((len(fasit),len(fasit)))
for i in range(len(fasit)):
    for j in range(len(fasit)):
        tall = np.abs(fasit[i]-fasit[j])
        distanceMatrSimple[i][j] = tall
    


pyexcel.save_as(array=distanceMatrSimple, dest_file_name="B11DistanceMatrix.xlsx")



N = np.arange(len(distanceMatrSimple))


Start = N[0]
Slutt = N[-1]


#range_bat = 36.6 #Baronen 
#range_bat = 41.48 #Baronessen
range_bat = 30.44 #Tidevind



#cost_nm = 0.272067 #Baronen 
#cost_nm = 0.19095 #Baronessen
cost_nm = 0.225 #MNOK/nm[range] TIDEVIND



A = [(i,j) for i in N for j in N if i<j] # All possible arcs between i and j, excluding when i == j
S = {(i,j): distanceMatrSimple[i,j] for i,j in A if j!=0 and distanceMatrSimple[i,j] <= range_bat } 

cost = [0]*len(N)#Cost of establishing charging in each port
cost[0] = 1.5
cost[1] = 1.5/7
cost[2] = 1.5/7
cost[3] = cost[1]
cost[4] = cost[2]
cost[5] = cost[1]
cost[6] = cost[2]
cost[7]  = cost[1]
cost[8] = cost[2]
cost[9] = cost[1]
cost[10] = cost[2]
cost[11] = cost[1]
cost[12] = cost[2]
cost[13] = cost[1]
cost[14] = cost[2]
cost[15] = 0



arcsData = []
for i,j in S:
    arcsData.append((i,j,S[i,j])) 



G.add_edges_from(S)

#GUROBI
#The optimization model 

m = gp.Model()
m.update()
x = m.addVars(S, vtype = GRB.BINARY, name = 'x') #declaring variables
#y = m.addVars(N,Bat_alt, vtype = GRB.BINARY, name = 'y')
u = m.addVars(N, vtype = GRB.BINARY, name = 'u')
w = m.addVars(S, vtype = GRB.BINARY, name = 'w')
#g = m.addVars(S, vtype = GRB.CONTINUOUS, name = 'I')

batCap = m.addVar(name = 'BatCap', vtype = GRB.CONTINUOUS)


m.modelSense = GRB.MINIMIZE #minimizing the Objective function

m.setObjective(gp.quicksum((cost[i]*u[i] for i in N )) + batCap*cost_nm/10   ) 

start_const = []
for a in arcsData:
    if a[0] == Start:
        start_const.append(a)

end_const = []
for a in arcsData:
    if a[1] == Slutt:
        end_const.append(a)


m.addConstrs(w[i,j]-x[i,j] == 0 for i,j in S)
m.addConstrs(w[i,j] >= x[i,j]+u[i]-1 for i,j in S)
m.addConstrs(w[i,j]<= x[i,j] for i,j in S)
m.addConstrs(w[i,j]<= u[i] for i,j in S)



m.addLConstr(gp.quicksum(x[a[0],a[1]] for a in start_const) == 1)
m.addLConstr(gp.quicksum(x[a[0],a[1]] for a in end_const) == 1) 
m.addConstrs((x[i,j]*S[i,j]<= batCap*0.8 for i,j in S), name='MaxConstraint')
m.addConstr(batCap <=range_bat)


m.addConstrs((u[i+1] - u[i+3])  == 0 for i in N[0:-4]) 
m.addConstr(u[Start] == u[Slutt])

for v in G.nodes:
    if v not in [Start,Slutt]:

        expr1 = gp.quicksum(x[i,v] for i in G.predecessors(v))

        expr2 = gp.quicksum(x[v,j] for j in G.successors(v))

        m.addLConstr(expr1 - expr2 == 0)


m.optimize()

print(m.display())

m.write('test.mps')
if m.Status != GRB.OPTIMAL:
    print('Optimization ended with status %d' % m.Status)
    sys.exit(0)

for v in m.getVars():
    print('%s %g' % (v.VarName, v.X))


print('The Objective value: %g' % m.ObjVal)


result_nodes = []
for v in m.getVars():

    if v.X == 1 and 'u' in v.VarName:
        result_nodes.append(v.varName)


mellomhold = []
for a in result_nodes:
    b = re.findall(r'\d+',a)
    mellomhold.extend(b)

results_vec = []
for b in mellomhold:
    if b.isdigit():
        results_vec.append(int(b))

results_vec = list(sorted(set(results_vec)))
print(results_vec)


battery_data = []
for v in m.getVars():
    if v.X > 0.5 and 'y' in v.VarName:
        battery_data.append(v.VarName)

mellomhold_bat = []
for a in battery_data:
    b = re.findall(r'\d+',a)
    mellomhold_bat.extend(b)

results_vec_y = []
for b in mellomhold_bat:
    if b.isdigit():
        results_vec_y.append(int(b))

battery_station = results_vec_y[0::2]
batter_choice = results_vec_y[1::2]

for i in range(len(battery_station)):
    print('Port %s will supply vessel with %s batteries' %(battery_station[i],batter_choice[i]) )
        

print('The nodes in the solution:',results_vec)



pos = nx.kamada_kawai_layout(G)


options = {'node_color': 'cyan','node_size': 200,'width': 0.5}


plt.figure()
nx.draw(G,pos, with_labels=True, **options)


plt.figure()
nx.draw(G,pos, with_labels=True, **options)
shortPath = [(results_vec[i],results_vec[i-1]) for i in range(len(results_vec)-1,0,-1)]
nx.draw_networkx_edges(G,pos,edgelist=shortPath,width=5,edge_color='red')




plt.show()

